/* libxsvg - Render XVG documents using the Xr library
 *
 * Copyright  2002 USC/Information Sciences Institute
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that the above copyright notice appear in all copies
 * and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of
 * Information Sciences Institute not be used in advertising or
 * publicity pertaining to distribution of the software without
 * specific, written prior permission.  Information Sciences Institute
 * makes no representations about the suitability of this software for
 * any purpose.  It is provided "as is" without express or implied
 * warranty.
 *
 * INFORMATION SCIENCES INSTITUTE DISCLAIMS ALL WARRANTIES WITH REGARD
 * TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL INFORMATION SCIENCES
 * INSTITUTE BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Carl Worth <cworth@isi.edu>
 */

#include <stdlib.h>
#include <string.h>

#include "xsvgint.h"
#include "math.h"

static svg_status_t
_xsvg_begin_group (void *closure, double opacity);

static svg_status_t
_xsvg_begin_element (void *closure);

static svg_status_t
_xsvg_end_element (void *closure);

static svg_status_t
_xsvg_end_group (void *closure, double opacity);

static svg_status_t
_xsvg_move_to (void *closure, double x, double y);

static svg_status_t
_xsvg_line_to (void *closure, double x, double y);

static svg_status_t
_xsvg_curve_to (void *closure,
		double x1, double y1,
		double x2, double y2,
		double x3, double y3);

static svg_status_t
_xsvg_quadratic_curve_to (void *closure,
			  double x1, double y1,
			  double x2, double y2);

static svg_status_t
_xsvg_arc_to (void *closure,
	      double	rx,
	      double	ry,
	      double	x_axis_rotation,
	      int	large_arc_flag,
	      int	sweep_flag,
	      double	x,
	      double	y);

static void
_svg_path_arc_segment (cairo_t *cr,
		       double xc, double yc,
		       double th0, double th1,
		       double rx, double ry, double x_axis_rotation);

static svg_status_t
_xsvg_close_path (void *closure);

static svg_status_t
_xsvg_set_color (void *closure, const svg_color_t *color);

static svg_status_t
_xsvg_set_fill_opacity (void *closure, double fill_opacity);

static svg_status_t
_xsvg_set_fill_paint (void *closure, const svg_paint_t *paint);

static svg_status_t
_xsvg_set_fill_rule (void *closure, svg_fill_rule_t fill_rule);

static svg_status_t
_xsvg_select_font (xsvg_t *xsvg);

static svg_status_t
_xsvg_set_font_family (void *closure, const char *family);

static svg_status_t
_xsvg_set_font_size (void *closure, double size);

static svg_status_t
_xsvg_set_font_style (void *closure, svg_font_style_t font_style);

static svg_status_t
_xsvg_set_font_weight (void *closure, unsigned int weight);

static svg_status_t
_xsvg_set_opacity (void *closure, double opacity);

static svg_status_t
_xsvg_set_stroke_dash_array (void *closure, double *dash, int num_dashes);

static svg_status_t
_xsvg_set_stroke_dash_offset (void *closure, svg_length_t *offset);

static svg_status_t
_xsvg_set_stroke_line_cap (void *closure, svg_stroke_line_cap_t line_cap);

static svg_status_t
_xsvg_set_stroke_line_join (void *closure, svg_stroke_line_join_t line_join);

static svg_status_t
_xsvg_set_stroke_miter_limit (void *closure, double limit);

static svg_status_t
_xsvg_set_stroke_opacity (void *closure, double stroke_opacity);

static svg_status_t
_xsvg_set_stroke_paint (void *closure, const svg_paint_t *paint);

static svg_status_t
_xsvg_set_stroke_width (void *closure, svg_length_t *width);

static svg_status_t
_xsvg_set_text_anchor (void *closure, svg_text_anchor_t text_anchor);

static svg_status_t
_xsvg_transform (void *closure,
		  double a, double b,
		  double c, double d,
		  double e, double f);

static svg_status_t
_xsvg_apply_view_box (void *closure,
		      svg_view_box_t view_box,
		      svg_length_t *width,
		      svg_length_t *height);

static svg_status_t
_xsvg_set_viewport_dimension (void *closure,
		    	      svg_length_t *width,
		    	      svg_length_t *height);

static svg_status_t
_xsvg_render_line (void *closure,
		   svg_length_t *x1_len, svg_length_t *y1_len,
		   svg_length_t *x2_len, svg_length_t *y2_len);

static svg_status_t
_xsvg_render_path (void *closure);

static svg_status_t
_xsvg_render_ellipse (void *closure,
		      svg_length_t *cx,
		      svg_length_t *cy,
		      svg_length_t *rx,
		      svg_length_t *ry);

static svg_status_t
_xsvg_render_rect (void *closure,
		   svg_length_t *x,
		   svg_length_t *y,
		   svg_length_t *width,
		   svg_length_t *height,
		   svg_length_t *rx,
		   svg_length_t *ry);

static svg_status_t
_xsvg_render_text (void *closure,
		   svg_length_t *x_len,
		   svg_length_t *y_len,
		   unsigned char *utf8);

static svg_status_t
_xsvg_render_image (void		*closure,
		    unsigned char	*data,
		    unsigned int	data_width,
		    unsigned int	data_height,
		    svg_length_t	*x,
		    svg_length_t	*y,
		    svg_length_t	*width,
		    svg_length_t	*height);

static svg_status_t
_cairo_status_to_svg_status (cairo_status_t xr_status);

static svg_status_t
_xsvg_push_state (xsvg_t *xsvg);

static svg_status_t
_xsvg_pop_state (xsvg_t *xsvg);

static svg_status_t
_xsvg_length_to_pixel (xsvg_t *xsvg, svg_length_t *length, double *pixel);

static svg_render_engine_t XSVG_RENDER_ENGINE = {
    /* hierarchy */
    _xsvg_begin_group,
    _xsvg_begin_element,
    _xsvg_end_element,
    _xsvg_end_group,
    /* path creation */
    _xsvg_move_to,
    _xsvg_line_to,
    _xsvg_curve_to,
    _xsvg_quadratic_curve_to,
    _xsvg_arc_to,
    _xsvg_close_path,
    /* style */
    _xsvg_set_color,
    _xsvg_set_fill_opacity,
    _xsvg_set_fill_paint,
    _xsvg_set_fill_rule,
    _xsvg_set_font_family,
    _xsvg_set_font_size,
    _xsvg_set_font_style,
    _xsvg_set_font_weight,
    _xsvg_set_opacity,
    _xsvg_set_stroke_dash_array,
    _xsvg_set_stroke_dash_offset,
    _xsvg_set_stroke_line_cap,
    _xsvg_set_stroke_line_join,
    _xsvg_set_stroke_miter_limit,
    _xsvg_set_stroke_opacity,
    _xsvg_set_stroke_paint,
    _xsvg_set_stroke_width,
    _xsvg_set_text_anchor,
    /* transform */
    _xsvg_transform,
    _xsvg_apply_view_box,
    _xsvg_set_viewport_dimension,
    /* drawing */
    _xsvg_render_line,
    _xsvg_render_path,
    _xsvg_render_ellipse,
    _xsvg_render_rect,
    _xsvg_render_text,
    _xsvg_render_image
};

xsvg_status_t
xsvg_create (xsvg_t **xsvg)
{
    xsvg_status_t status;

    *xsvg = malloc (sizeof (xsvg_t));
    if (*xsvg == NULL) {
	return XSVG_STATUS_NO_MEMORY;
    }

    (*xsvg)->cr = NULL;
    (*xsvg)->state = NULL;
    (*xsvg)->viewport_width = 450;
    (*xsvg)->viewport_height = 450;
 
    status = svg_create (&(*xsvg)->svg);
    if (status)
	return status;

    return XSVG_STATUS_SUCCESS;
}

xsvg_status_t
xsvg_destroy (xsvg_t *xsvg)
{
    xsvg_status_t status;

    status = svg_destroy (xsvg->svg);

    free (xsvg);

    return status;
}

xsvg_status_t
xsvg_parse_file (xsvg_t *xsvg, const char *file_name)
{
    return svg_parse_file (xsvg->svg, file_name);
}

xsvg_status_t
xsvg_parse_buffer (xsvg_t *xsvg, const char *buf, size_t count)
{
    return svg_parse_buffer (xsvg->svg, buf, count);
}

xsvg_status_t
xsvg_parse_chunk_begin (xsvg_t *xsvg)
{
    return svg_parse_chunk_begin (xsvg->svg);
}

xsvg_status_t
xsvg_parse_chunk (xsvg_t *xsvg, const char *buf, size_t count)
{
    return svg_parse_chunk (xsvg->svg, buf, count);
}

xsvg_status_t
xsvg_parse_chunk_end (xsvg_t *xsvg)
{
    return svg_parse_chunk_end (xsvg->svg);
}

xsvg_status_t
xsvg_render (xsvg_t *xsvg, cairo_t *cr)
{
    xsvg->cr = cr;
    return svg_render (xsvg->svg, &XSVG_RENDER_ENGINE, xsvg);
}

static svg_status_t
_xsvg_set_viewport_dimension (void *closure,
		    	      svg_length_t *width,
		    	      svg_length_t *height)
{
    xsvg_t *xsvg = closure;
    double vwidth, vheight;

    _xsvg_length_to_pixel (xsvg, width, &vwidth);
    _xsvg_length_to_pixel (xsvg, height, &vheight);

    /* TODO : change xsvg->state->viewport_width to type double
       if that accuracy is needed. */
    xsvg->state->viewport_width  = (unsigned int)vwidth;
    xsvg->state->viewport_height = (unsigned int)vheight;

    return XSVG_STATUS_SUCCESS;
}

xsvg_status_t
xsvg_set_viewport_dimension (xsvg_t *xsvg, unsigned int width, unsigned int height)
{
    xsvg->viewport_width = width;
    xsvg->viewport_height = height;

    return XSVG_STATUS_SUCCESS;
}

void
xsvg_get_size (xsvg_t *xsvg, int *width, int *height)
{
    svg_get_size (xsvg->svg, width, height);
}

static svg_status_t
_xsvg_begin_group (void *closure, double opacity)
{
    xsvg_t *xsvg = closure;

    cairo_save (xsvg->cr);

    if (opacity != 1.0) {
	xsvg->state->child_surface = cairo_surface_create_similar (cairo_current_target_surface (xsvg->cr),
							    CAIRO_FORMAT_ARGB32,
							    xsvg->viewport_width,
							    xsvg->viewport_height);
	cairo_set_target_surface (xsvg->cr, xsvg->state->child_surface);
    }

    _xsvg_push_state (xsvg);

    return _cairo_status_to_svg_status (cairo_status (xsvg->cr));
}

/* XXX: begin_element could be made more efficient in that no extra
   group is needed if there is only one element in a group */
static svg_status_t
_xsvg_begin_element (void *closure)
{
    xsvg_t *xsvg = closure;

    cairo_save (xsvg->cr);

    _xsvg_push_state (xsvg);

    return _cairo_status_to_svg_status (cairo_status (xsvg->cr));
}

static svg_status_t
_xsvg_end_element (void *closure)
{
    xsvg_t *xsvg = closure;

    _xsvg_pop_state (xsvg);

    cairo_restore (xsvg->cr);

    return _cairo_status_to_svg_status (cairo_status (xsvg->cr));
}

static svg_status_t
_xsvg_end_group (void *closure, double opacity)
{
    xsvg_t *xsvg = closure;

    _xsvg_pop_state (xsvg);

    cairo_restore (xsvg->cr);

    if (opacity != 1.0) {
	cairo_save (xsvg->cr);
	/* Put the DefaultMatrix in place, (XXX: do we want
           IdentityMatrix here?), or else the current transformation
           will have been applied twice. I'm still a bit fuzzy on
           whether this is the correct approach or if it's masking
           some other problem. */
	cairo_default_matrix (xsvg->cr);
	cairo_set_alpha (xsvg->cr, opacity);
	cairo_show_surface (xsvg->cr, xsvg->state->child_surface,
		       xsvg->viewport_width, xsvg->viewport_height);
	cairo_restore (xsvg->cr);
	cairo_surface_destroy (xsvg->state->child_surface);
	xsvg->state->child_surface = NULL;
    }

    return _cairo_status_to_svg_status (cairo_status (xsvg->cr));
}

static svg_status_t
_xsvg_move_to (void *closure, double x, double y)
{
    xsvg_t *xsvg = closure;

    cairo_move_to (xsvg->cr, x, y);

    return _cairo_status_to_svg_status (cairo_status (xsvg->cr));
}

static svg_status_t
_xsvg_line_to (void *closure, double x, double y)
{
    xsvg_t *xsvg = closure;

    cairo_line_to (xsvg->cr, x, y);

    return _cairo_status_to_svg_status (cairo_status (xsvg->cr));
}

static svg_status_t
_xsvg_curve_to (void *closure,
		double x1, double y1,
		double x2, double y2,
		double x3, double y3)
{
    xsvg_t *xsvg = closure;

    cairo_curve_to (xsvg->cr,
		    x1, y1,
		    x2, y2,
		    x3, y3);

    return _cairo_status_to_svg_status (cairo_status (xsvg->cr));
}

static svg_status_t
_xsvg_quadratic_curve_to (void *closure,
			  double x1, double y1,
			  double x2, double y2)
{
    xsvg_t *xsvg = closure;
    double x, y;

    cairo_current_point (xsvg->cr, &x, &y);

    cairo_curve_to (xsvg->cr,
		    x  + 2.0/3.0 * (x1 - x),  y  + 2.0/3.0 * (y1 - y),
		    x2 + 2.0/3.0 * (x1 - x2), y2 + 2.0/3.0 * (y1 - y2),
		    x2, y2);

    return _cairo_status_to_svg_status (cairo_status (xsvg->cr));
}

static svg_status_t
_xsvg_close_path (void *closure)
{
    xsvg_t *xsvg = closure;

    cairo_close_path (xsvg->cr);

    return _cairo_status_to_svg_status (cairo_status (xsvg->cr));
}

static svg_status_t
_xsvg_set_color (void *closure, const svg_color_t *color)
{
    xsvg_t *xsvg = closure;

    if (color->is_current_color)
	return SVG_STATUS_SUCCESS;

    cairo_set_rgb_color (xsvg->cr,
			 svg_color_get_red   (color) / 255.0,
			 svg_color_get_green (color) / 255.0,
			 svg_color_get_blue  (color) / 255.0);

    return _cairo_status_to_svg_status (cairo_status (xsvg->cr));
}

static svg_status_t
_xsvg_set_alpha (xsvg_t *xsvg, double alpha)
{
    cairo_set_alpha (xsvg->cr, alpha);

    return _cairo_status_to_svg_status (cairo_status (xsvg->cr));
}

static svg_status_t
_xsvg_set_gradient (xsvg_t *xsvg, svg_gradient_t *gradient)
{
    switch (gradient->type) {
      case SVG_GRADIENT_LINEAR:
      {
	  int i;
	  double x1, y1, x2, y2;
	  double scale;
	  cairo_surface_t *pattern;
	  cairo_matrix_t *matrix;
	  svg_gradient_stop_t *stop;

	  cairo_save (xsvg->cr);
	  cairo_identity_matrix (xsvg->cr);
	  cairo_new_path (xsvg->cr);

	  pattern = cairo_surface_create_similar (cairo_current_target_surface (xsvg->cr),
						  CAIRO_FORMAT_ARGB32,
						  gradient->num_stops,
						  1);
	  cairo_set_target_surface (xsvg->cr, pattern);
	  
	  /* XXX: This is a major hack right now... I'm ignoring stop->offset. */
	  for (i=0; i < gradient->num_stops; i++) {
	      stop = &gradient->stops[i];
	      cairo_rectangle (xsvg->cr, i, 0, 1, 1);
	      cairo_set_rgb_color (xsvg->cr,
				   svg_color_get_red (&stop->color) / 255.0,
				   svg_color_get_green (&stop->color) / 255.0,
				   svg_color_get_blue (&stop->color) / 255.0);
	      cairo_set_alpha (xsvg->cr, stop->opacity);
	      cairo_fill (xsvg->cr);
	  }
	  
	  cairo_restore (xsvg->cr);

	  /* XXX: Not yet handling gradients with percentage units in
             terms of the current object (gradientUnits="objectBoundingBox").  

	     To support those, we will need to be able to compute the
	     bounds of the current object.
	  */
	  _xsvg_length_to_pixel (xsvg, &gradient->u.linear.x1, &x1);
	  _xsvg_length_to_pixel (xsvg, &gradient->u.linear.y1, &y1);
	  _xsvg_length_to_pixel (xsvg, &gradient->u.linear.x2, &x2);
	  _xsvg_length_to_pixel (xsvg, &gradient->u.linear.y2, &y2);

	  scale = sqrt ((x2 - x1)*(x2 - x1) + (y2 - y1)*(y2 - y1));
	  matrix = cairo_matrix_create ();
	  /* XXX: The horizontal scale factor here must be sufficient
             for covering the object completely. So we'll need some
             estimate on the bounds of the object. For now, I'll just
             throw in an arbitrary hard-coded scale factor. */
	  cairo_matrix_scale (matrix,
			      (gradient->num_stops - 1) / scale,
			      1.0/65536.0);
	  cairo_matrix_rotate (matrix,
			       -atan2 (y2-y1, x2-x1));
	  cairo_matrix_translate (matrix,
				  -x1, -y1);
	  cairo_surface_set_matrix (pattern, matrix);
	  cairo_surface_set_filter (pattern, CAIRO_FILTER_BILINEAR);
	  cairo_matrix_destroy (matrix);

	  cairo_move_to (xsvg->cr, 0, 0);
	  cairo_set_pattern (xsvg->cr, pattern);
	  
	  cairo_surface_destroy (pattern);
      }
      break;
      case SVG_GRADIENT_RADIAL:
	/* XXX: NYI */
	break;
    }

    return SVG_STATUS_SUCCESS;
}

static svg_status_t
_xsvg_set_pattern (xsvg_t *xsvg, svg_element_t *pattern_element)
{
    svg_pattern_t *pattern = svg_element_pattern (pattern_element);
    cairo_surface_t *pattern_surface;
    double x_px, y_px, width_px, height_px;

    _xsvg_length_to_pixel (xsvg, &pattern->x, &x_px);
    _xsvg_length_to_pixel (xsvg, &pattern->y, &y_px);
    _xsvg_length_to_pixel (xsvg, &pattern->width, &width_px);
    _xsvg_length_to_pixel (xsvg, &pattern->height, &height_px);

    cairo_save (xsvg->cr);

    pattern_surface = cairo_surface_create_similar (cairo_current_target_surface (xsvg->cr),
						    CAIRO_FORMAT_ARGB32,
						    (int) (width_px + 0.5),
						    (int) (height_px + 0.5));

    cairo_set_target_surface (xsvg->cr, pattern_surface);

    /* XXX: The pattern support is a real mess right now. This hack
       prevents a case of infinite recursion which is just one
       artifact of the current mess. Part of the solution involves
       restoring the graphics state of the associated pattern before
       drawing it. */
    xsvg->state->in_set_pattern_HACK = 1;
    svg_element_render (pattern->group_element, &XSVG_RENDER_ENGINE, xsvg);
    xsvg->state->in_set_pattern_HACK = 0;

    cairo_restore (xsvg->cr);

    cairo_surface_set_repeat (pattern_surface, 1);
    cairo_set_pattern (xsvg->cr, pattern_surface);

    cairo_surface_destroy (pattern_surface);

    return SVG_STATUS_SUCCESS;
}

static svg_status_t
_xsvg_set_paint_and_opacity (xsvg_t *xsvg, svg_paint_t *paint, double opacity)
{
    svg_status_t status;

    switch (paint->type) {
    case SVG_PAINT_TYPE_NONE:
	break;
    case SVG_PAINT_TYPE_COLOR:
	status = _xsvg_set_color (xsvg, &paint->p.color);
	if (status)
	    return status;
	status = _xsvg_set_alpha (xsvg, opacity * xsvg->state->opacity);
	if (status)
	    return status;
	break;
    case SVG_PAINT_TYPE_GRADIENT:
	status = _xsvg_set_gradient (xsvg, paint->p.gradient);
	if (status)
	    return status;
	break;
    case SVG_PAINT_TYPE_PATTERN:
	status = _xsvg_set_pattern (xsvg, paint->p.pattern_element);
	if (status)
	    return status;
	break;
    }

    return SVG_STATUS_SUCCESS;
}

static svg_status_t
_xsvg_set_fill_opacity (void *closure, double fill_opacity)
{
    xsvg_t *xsvg = closure;

    xsvg->state->fill_opacity = fill_opacity;

    return _cairo_status_to_svg_status (cairo_status (xsvg->cr));    
}

static svg_status_t
_xsvg_set_fill_paint (void *closure, const svg_paint_t *paint)
{
    xsvg_t *xsvg = closure;

    xsvg->state->fill_paint = *paint;

    return SVG_STATUS_SUCCESS;
}

static svg_status_t
_xsvg_set_fill_rule (void *closure, svg_fill_rule_t fill_rule)
{
    xsvg_t *xsvg = closure;

    switch (fill_rule) {
    case SVG_FILL_RULE_NONZERO:
	cairo_set_fill_rule (xsvg->cr, CAIRO_FILL_RULE_WINDING);
	break;
    case SVG_FILL_RULE_EVEN_ODD:
	cairo_set_fill_rule (xsvg->cr, CAIRO_FILL_RULE_EVEN_ODD);
	break;
    }

    return _cairo_status_to_svg_status (cairo_status (xsvg->cr));    
}

static svg_status_t
_xsvg_select_font (xsvg_t *xsvg)
{
    char *family = xsvg->state->font_family;
    unsigned int font_weight = xsvg->state->font_weight;
    cairo_font_weight_t weight;
    svg_font_style_t font_style = xsvg->state->font_style;
    cairo_font_slant_t slant;

    if (! xsvg->state->font_dirty)
	return SVG_STATUS_SUCCESS;

    if (font_weight >= 700)
	weight = CAIRO_FONT_WEIGHT_BOLD;
    else
	weight = CAIRO_FONT_WEIGHT_NORMAL;

    switch (font_style) {
    case SVG_FONT_STYLE_ITALIC:
	slant = CAIRO_FONT_SLANT_ITALIC;
	break;
    case SVG_FONT_STYLE_OBLIQUE:
	slant = CAIRO_FONT_SLANT_OBLIQUE;
	break;
    default:
	slant = CAIRO_FONT_SLANT_NORMAL;
	break;
    }

    cairo_select_font (xsvg->cr, family, weight, slant);
    cairo_scale_font (xsvg->cr, xsvg->state->font_size);
    xsvg->state->font_dirty = 0;

    return _cairo_status_to_svg_status (cairo_status (xsvg->cr));
}

static svg_status_t
_xsvg_set_font_family (void *closure, const char *family)
{
    xsvg_t *xsvg = closure;

    if (xsvg->state->font_family)
	free (xsvg->state->font_family);

    xsvg->state->font_family = strdup (family);
    xsvg->state->font_dirty = 1;

    return _cairo_status_to_svg_status (cairo_status (xsvg->cr));
}

static svg_status_t
_xsvg_set_font_size (void *closure, double size)
{
    xsvg_t *xsvg = closure;

    xsvg->state->font_size = size;
    xsvg->state->font_dirty = 1;

    return _cairo_status_to_svg_status (cairo_status (xsvg->cr));
}

static svg_status_t
_xsvg_set_font_style (void *closure, svg_font_style_t font_style)
{
    xsvg_t *xsvg = closure;

    xsvg->state->font_style = font_style;
    xsvg->state->font_dirty = 1;

    return _cairo_status_to_svg_status (cairo_status (xsvg->cr));
}

static svg_status_t
_xsvg_set_font_weight (void *closure, unsigned int font_weight)
{
    xsvg_t *xsvg = closure;

    xsvg->state->font_weight = font_weight;
    xsvg->state->font_dirty = 1;

    return _cairo_status_to_svg_status (cairo_status (xsvg->cr));
}

static svg_status_t
_xsvg_set_opacity (void *closure, double opacity)
{
    xsvg_t *xsvg = closure;

    xsvg->state->opacity = opacity;

    return _cairo_status_to_svg_status (cairo_status (xsvg->cr));    
}

static svg_status_t
_xsvg_set_stroke_dash_array (void *closure, double *dash, int num_dashes)
{
    xsvg_t *xsvg = closure;

    free (xsvg->state->dash);
    xsvg->state->dash = NULL;

    xsvg->state->num_dashes = num_dashes;

    if (xsvg->state->num_dashes) {
	xsvg->state->dash = malloc(xsvg->state->num_dashes * sizeof(double));
	if (xsvg->state->dash == NULL)
	    return SVG_STATUS_NO_MEMORY;

	memcpy(xsvg->state->dash, dash, xsvg->state->num_dashes * sizeof(double));

	cairo_set_dash (xsvg->cr, xsvg->state->dash, xsvg->state->num_dashes, xsvg->state->dash_offset);
    }

    return _cairo_status_to_svg_status (cairo_status (xsvg->cr));
}

static svg_status_t
_xsvg_set_stroke_dash_offset (void *closure, svg_length_t *offset_len)
{
    xsvg_t *xsvg = closure;
    double offset;

    _xsvg_length_to_pixel (xsvg, offset_len, &offset);

    xsvg->state->dash_offset = offset;

    if (xsvg->state->num_dashes)
	cairo_set_dash (xsvg->cr, xsvg->state->dash, xsvg->state->num_dashes, xsvg->state->dash_offset); 
    return _cairo_status_to_svg_status (cairo_status (xsvg->cr));
}

static svg_status_t
_xsvg_set_stroke_line_cap (void *closure, svg_stroke_line_cap_t line_cap)
{
    xsvg_t *xsvg = closure;

    switch (line_cap) {
    case SVG_STROKE_LINE_CAP_BUTT:
	cairo_set_line_cap (xsvg->cr, CAIRO_LINE_CAP_BUTT);
	break;
    case SVG_STROKE_LINE_CAP_ROUND:
	cairo_set_line_cap (xsvg->cr, CAIRO_LINE_CAP_ROUND);
	break;
    case SVG_STROKE_LINE_CAP_SQUARE:
	cairo_set_line_cap (xsvg->cr, CAIRO_LINE_CAP_SQUARE);
	break;
    }

    return _cairo_status_to_svg_status (cairo_status (xsvg->cr));
}

static svg_status_t
_xsvg_set_stroke_line_join (void *closure, svg_stroke_line_join_t line_join)
{
    xsvg_t *xsvg = closure;

    switch (line_join) {
    case SVG_STROKE_LINE_JOIN_MITER:
	cairo_set_line_join (xsvg->cr, CAIRO_LINE_JOIN_MITER);
	break;
    case SVG_STROKE_LINE_JOIN_ROUND:
	cairo_set_line_join (xsvg->cr, CAIRO_LINE_JOIN_ROUND);
	break;
    case SVG_STROKE_LINE_JOIN_BEVEL:
	cairo_set_line_join (xsvg->cr, CAIRO_LINE_JOIN_BEVEL);
	break;
    }

    return _cairo_status_to_svg_status (cairo_status (xsvg->cr));
}

static svg_status_t
_xsvg_set_stroke_miter_limit (void *closure, double limit)
{
    xsvg_t *xsvg = closure;

    cairo_set_miter_limit (xsvg->cr, limit);

    return _cairo_status_to_svg_status (cairo_status (xsvg->cr));    
}

static svg_status_t
_xsvg_set_stroke_opacity (void *closure, double stroke_opacity)
{
    xsvg_t *xsvg = closure;

    xsvg->state->stroke_opacity = stroke_opacity;

    return _cairo_status_to_svg_status (cairo_status (xsvg->cr));    
}

static svg_status_t
_xsvg_set_stroke_paint (void *closure, const svg_paint_t *paint)
{
    xsvg_t *xsvg = closure;

    xsvg->state->stroke_paint = *paint;

    return SVG_STATUS_SUCCESS;
}

static svg_status_t
_xsvg_set_stroke_width (void *closure, svg_length_t *width_len)
{
    xsvg_t *xsvg = closure;
    double width;

    _xsvg_length_to_pixel (xsvg, width_len, &width);

    cairo_set_line_width (xsvg->cr, width);

    return _cairo_status_to_svg_status (cairo_status (xsvg->cr));    
}

static svg_status_t
_xsvg_set_text_anchor (void *closure, svg_text_anchor_t text_anchor)
{
    xsvg_t *xsvg = closure;

    xsvg->state->text_anchor = text_anchor;

    return SVG_STATUS_SUCCESS;
}

static svg_status_t
_xsvg_transform (void *closure,
		  double a, double b,
		  double c, double d,
		  double e, double f)
{
    xsvg_t *xsvg = closure;
    cairo_matrix_t *matrix;

    matrix = cairo_matrix_create ();
    cairo_matrix_set_affine (matrix, a, b, c, d, e, f);

    cairo_concat_matrix (xsvg->cr, matrix);

    cairo_matrix_destroy (matrix);

    return _cairo_status_to_svg_status (cairo_status (xsvg->cr));    
}

static svg_status_t
_xsvg_render_line (void *closure,
		   svg_length_t *x1_len, svg_length_t *y1_len,
		   svg_length_t *x2_len, svg_length_t *y2_len)
{
    xsvg_t *xsvg = closure;
    svg_status_t status;
    double x1, y1, x2, y2;

    _xsvg_length_to_pixel (xsvg, x1_len, &x1);
    _xsvg_length_to_pixel (xsvg, y1_len, &y1);
    _xsvg_length_to_pixel (xsvg, x2_len, &x2);
    _xsvg_length_to_pixel (xsvg, y2_len, &y2);

    status = _xsvg_move_to (xsvg, x1, y1);
    if (status)
	return status;

    status = _xsvg_line_to (xsvg, x2, y2);
    if (status)
	return status;

    status = _xsvg_render_path (xsvg);
    if (status)
	return status;

    return SVG_STATUS_SUCCESS;

}

static svg_status_t
_xsvg_render_path (void *closure)
{
    xsvg_t *xsvg = closure;
    svg_paint_t *fill_paint, *stroke_paint;

    fill_paint = &xsvg->state->fill_paint;
    stroke_paint = &xsvg->state->stroke_paint;

    if (fill_paint->type
	&& (fill_paint->type != SVG_PAINT_TYPE_PATTERN || !xsvg->state->in_set_pattern_HACK)) {
	if (stroke_paint->type)
	    cairo_save (xsvg->cr);
	_xsvg_set_paint_and_opacity (xsvg, fill_paint,
				     xsvg->state->fill_opacity);
	cairo_fill (xsvg->cr);
	if (stroke_paint->type)
	    cairo_restore (xsvg->cr);
    }

    if (stroke_paint->type
	&& (stroke_paint->type != SVG_PAINT_TYPE_PATTERN || !xsvg->state->in_set_pattern_HACK)) {
	_xsvg_set_paint_and_opacity (xsvg, stroke_paint,
				     xsvg->state->stroke_opacity);
	cairo_stroke (xsvg->cr);
    }

    return _cairo_status_to_svg_status (cairo_status (xsvg->cr));
}

static svg_status_t
_xsvg_render_ellipse (void *closure,
		      svg_length_t *cx_len,
		      svg_length_t *cy_len,
		      svg_length_t *rx_len,
		      svg_length_t *ry_len)
{
    xsvg_t *xsvg = closure;
    cairo_matrix_t *matrix;

    double cx, cy, rx, ry;

    _xsvg_length_to_pixel (xsvg, cx_len, &cx);
    _xsvg_length_to_pixel (xsvg, cy_len, &cy);
    _xsvg_length_to_pixel (xsvg, rx_len, &rx);
    _xsvg_length_to_pixel (xsvg, ry_len, &ry);

    matrix = cairo_matrix_create ();
    cairo_current_matrix (xsvg->cr, matrix);

    cairo_translate (xsvg->cr, cx, cy);
    cairo_scale (xsvg->cr, 1.0, ry / rx);
    cairo_move_to (xsvg->cr, rx, 0.0);
    cairo_arc (xsvg->cr, 0, 0, rx, 0, 2 * M_PI);
    cairo_close_path (xsvg->cr);

    cairo_set_matrix (xsvg->cr, matrix);
    cairo_matrix_destroy (matrix);

     _xsvg_render_path (xsvg);

    return _cairo_status_to_svg_status (cairo_status (xsvg->cr));
}

static svg_status_t
_xsvg_render_rect (void *closure,
		   svg_length_t *x_len,
		   svg_length_t *y_len,
		   svg_length_t *width_len,
		   svg_length_t *height_len,
		   svg_length_t *rx_len,
		   svg_length_t *ry_len)
{
    xsvg_t *xsvg = closure;

    double x, y, width, height, rx, ry;
 
    _xsvg_length_to_pixel (xsvg, x_len, &x);
    _xsvg_length_to_pixel (xsvg, y_len, &y);
    _xsvg_length_to_pixel (xsvg, width_len, &width);
    _xsvg_length_to_pixel (xsvg, height_len, &height);
    _xsvg_length_to_pixel (xsvg, rx_len, &rx);
    _xsvg_length_to_pixel (xsvg, ry_len, &ry);
 
    if (rx > width / 2.0)
	rx = width / 2.0;
    if (ry > height / 2.0)
	ry = height / 2.0;

    if (rx > 0 || ry > 0)
    {
	_xsvg_move_to (xsvg, x + rx, y);
	_xsvg_line_to (xsvg, x + width - rx, y);
	_xsvg_arc_to  (xsvg, rx, ry, 0, 0, 1, x + width, y + ry);
	_xsvg_line_to (xsvg, x + width, y + height - ry);
	_xsvg_arc_to  (xsvg, rx, ry, 0, 0, 1, x + width - rx, y + height);
	_xsvg_line_to (xsvg, x + rx, y + height);
	_xsvg_arc_to  (xsvg, rx, ry, 0, 0, 1, x, y + height - ry);
	_xsvg_line_to (xsvg, x, y + ry);
	_xsvg_arc_to  (xsvg, rx, ry, 0, 0, 1, x + rx, y);
    }
    else
    {
	_xsvg_move_to (xsvg, x, y);
	_xsvg_line_to (xsvg, x + width, y);
	_xsvg_line_to (xsvg, x + width, y + height);
	_xsvg_line_to (xsvg, x, y + height);
    }
    _xsvg_close_path (xsvg);

    _xsvg_render_path (xsvg);

    return _cairo_status_to_svg_status (cairo_status (xsvg->cr));
}

static svg_status_t
_xsvg_render_text (void *closure,
		   svg_length_t *x_len,
		   svg_length_t *y_len,
		   unsigned char *utf8)
{
    xsvg_t *xsvg = closure;
    svg_paint_t *paint;
    double x, y;
    svg_status_t status;

    _xsvg_length_to_pixel (xsvg, x_len, &x);
    _xsvg_length_to_pixel (xsvg, y_len, &y);

    status = _xsvg_move_to (xsvg, x, y);
    if (status)
	return status;

    paint = &xsvg->state->fill_paint;
    if (paint->type == SVG_PAINT_TYPE_COLOR) {
	_xsvg_set_color (xsvg, &paint->p.color);
	cairo_set_alpha (xsvg->cr, xsvg->state->fill_opacity * xsvg->state->opacity);
    }

/* XXX: Temporarily disable since cairo_text_extents is currently not working

    if (xsvg->state->text_anchor != SVG_TEXT_ANCHOR_START) {
	double x, y, width, height, dx, dy;
	cairo_text_extents (xsvg->cr, utf8, &x, &y, &width, &height, &dx, &dy);
	if (xsvg->state->text_anchor == SVG_TEXT_ANCHOR_END)
	    cairo_rel_move_to (xsvg->cr, -dx, -dy);
	else if (xsvg->state->text_anchor == SVG_TEXT_ANCHOR_MIDDLE)
	    cairo_rel_move_to (xsvg->cr, -dx / 2, -dy / 2);
    }
*/

    _xsvg_select_font (xsvg);
    cairo_show_text (xsvg->cr, utf8);

    return _cairo_status_to_svg_status (cairo_status (xsvg->cr));
}

static svg_status_t
_xsvg_render_image (void		*closure,
		    unsigned char	*data,
		    unsigned int	data_width,
		    unsigned int	data_height,
		    svg_length_t	*x_len,
		    svg_length_t	*y_len,
		    svg_length_t	*width_len,
		    svg_length_t	*height_len)
{
    xsvg_t *xsvg = closure;
    cairo_surface_t *surface;
    double x, y, width, height;

    _xsvg_length_to_pixel (xsvg, x_len, &x);
    _xsvg_length_to_pixel (xsvg, y_len, &y);
    _xsvg_length_to_pixel (xsvg, width_len, &width);
    _xsvg_length_to_pixel (xsvg, height_len, &height);

    surface = cairo_surface_create_for_image (data, CAIRO_FORMAT_ARGB32,
					      data_width, data_height, data_width *4);
    cairo_translate (xsvg->cr, x, y);
    cairo_scale (xsvg->cr, width / data_width, height / data_height);
    cairo_set_alpha (xsvg->cr, xsvg->state->opacity);

    cairo_show_surface (xsvg->cr,
			surface,
			data_width, data_height);

    cairo_surface_destroy (surface);

    return _cairo_status_to_svg_status (cairo_status (xsvg->cr));
}

static svg_status_t
_cairo_status_to_svg_status (cairo_status_t xr_status)
{
    switch (xr_status) {
    case CAIRO_STATUS_NO_MEMORY:
	return SVG_STATUS_NO_MEMORY;
    case CAIRO_STATUS_SUCCESS:
    default:
	return SVG_STATUS_SUCCESS;
    }
}

static svg_status_t
_xsvg_push_state (xsvg_t *xsvg)
{
    if (!xsvg->state)
    {
	xsvg->state = _xsvg_state_push (xsvg->state);
	xsvg->state->viewport_width = xsvg->viewport_width;
	xsvg->state->viewport_height = xsvg->viewport_height;
    }
    else
	xsvg->state = _xsvg_state_push (xsvg->state);
    if (xsvg->state == NULL)
	return SVG_STATUS_NO_MEMORY;

    return SVG_STATUS_SUCCESS;
}

static svg_status_t
_xsvg_pop_state (xsvg_t *xsvg)
{
    xsvg->state = _xsvg_state_pop (xsvg->state);

    return SVG_STATUS_SUCCESS;
}

#define	DPI	100.0 // This is NOT what we want. How to reach global DPI? (Rob)

static svg_status_t
_xsvg_length_to_pixel (xsvg_t * xsvg, svg_length_t *length, double *pixel)
{
    double width, height;

    switch (length->unit) {
    case SVG_LENGTH_UNIT_PX:
	*pixel = length->value;
	break;
    case SVG_LENGTH_UNIT_CM:
	*pixel = (length->value / 2.54) * DPI;
	break;
    case SVG_LENGTH_UNIT_MM:
	*pixel = (length->value / 25.4) * DPI;
	break;
    case SVG_LENGTH_UNIT_IN:
	*pixel = length->value * DPI;
	break;
    case SVG_LENGTH_UNIT_PT:
	*pixel = (length->value / 72.0) * DPI;
	break;
    case SVG_LENGTH_UNIT_PC:
	*pixel = (length->value / 6.0) * DPI;
	break;
    case SVG_LENGTH_UNIT_EM:
	*pixel = length->value * xsvg->state->font_size;
	break;
    case SVG_LENGTH_UNIT_EX:
	*pixel = length->value * xsvg->state->font_size / 2.0;
	break;
    case SVG_LENGTH_UNIT_PCT:
	/* TODO : we usually but not always want to compute against target window dimension.
	          Make it so we also computes against inner <svg> elements, maybe by storing a
		  width/height pair in the state? */
	width = xsvg->state->viewport_width;
	height = xsvg->state->viewport_height;
	if (length->orientation == SVG_LENGTH_ORIENTATION_HORIZONTAL)
	    *pixel = (length->value / 100.0) * width;
	else if (length->orientation == SVG_LENGTH_ORIENTATION_VERTICAL)
	    *pixel = (length->value / 100.0) * height;
	else
	    *pixel = (length->value / 100.0) * sqrt(pow(width, 2) + pow(height, 2)) * sqrt(2);
	break;
    default:
	*pixel = length->value;
    }

    return SVG_STATUS_SUCCESS;
}

static svg_status_t
_xsvg_apply_view_box (void *closure,
		      svg_view_box_t view_box,
		      svg_length_t *width,
		      svg_length_t *height)
{
    xsvg_t *xsvg = closure;
    double vpar, svgar;
    double logic_width, logic_height;
    double logic_x, logic_y;
    double phys_width, phys_height;
    _xsvg_length_to_pixel (xsvg, width, &phys_width);
    _xsvg_length_to_pixel (xsvg, height, &phys_height);

    vpar = view_box.box.width / view_box.box.height;
    svgar = phys_width / phys_height;
    logic_x = view_box.box.x;
    logic_y = view_box.box.y;
    logic_width = view_box.box.width;
    logic_height = view_box.box.height;

    if (view_box.aspect_ratio == SVG_PRESERVE_ASPECT_RATIO_NONE)
    {
	cairo_scale (xsvg->cr,
		     phys_width / logic_width,
		     phys_height / logic_height);
	cairo_translate (xsvg->cr, -logic_x, -logic_y);
    }
    else if ((vpar < svgar && view_box.meet_or_slice == SVG_MEET_OR_SLICE_MEET) ||
    	     (vpar >= svgar && view_box.meet_or_slice == SVG_MEET_OR_SLICE_SLICE))
    {
	cairo_scale (xsvg->cr, phys_height / logic_height, phys_height / logic_height);

	if (view_box.aspect_ratio == SVG_PRESERVE_ASPECT_RATIO_XMINYMIN ||
	    view_box.aspect_ratio == SVG_PRESERVE_ASPECT_RATIO_XMINYMID ||
	    view_box.aspect_ratio == SVG_PRESERVE_ASPECT_RATIO_XMINYMAX)
	    cairo_translate (xsvg->cr, -logic_x, -logic_y);
	else if(view_box.aspect_ratio == SVG_PRESERVE_ASPECT_RATIO_XMIDYMIN ||
		view_box.aspect_ratio == SVG_PRESERVE_ASPECT_RATIO_XMIDYMID ||
		view_box.aspect_ratio == SVG_PRESERVE_ASPECT_RATIO_XMIDYMAX)
            cairo_translate (xsvg->cr,
			 -logic_x - (logic_width - phys_width * logic_height / phys_height) / 2,
			 -logic_y);
	else
            cairo_translate (xsvg->cr,
			 -logic_x - (logic_width - phys_width * logic_height / phys_height),
			 -logic_y);
    }
    else
    {
	cairo_scale (xsvg->cr, phys_width / logic_width, phys_width / logic_width);

	if (view_box.aspect_ratio == SVG_PRESERVE_ASPECT_RATIO_XMINYMIN ||
	    view_box.aspect_ratio == SVG_PRESERVE_ASPECT_RATIO_XMIDYMIN ||
	    view_box.aspect_ratio == SVG_PRESERVE_ASPECT_RATIO_XMAXYMIN)
	    cairo_translate (xsvg->cr, -logic_x, -logic_y);
	else if(view_box.aspect_ratio == SVG_PRESERVE_ASPECT_RATIO_XMINYMID ||
		view_box.aspect_ratio == SVG_PRESERVE_ASPECT_RATIO_XMIDYMID ||
		view_box.aspect_ratio == SVG_PRESERVE_ASPECT_RATIO_XMAXYMID)
	    cairo_translate (xsvg->cr,
			 -logic_x,
			 -logic_y - (logic_height - phys_height * logic_width / phys_width) / 2);
	else
	    cairo_translate (xsvg->cr,
			 -logic_x,
			 -logic_y - (logic_height - phys_height * logic_width / phys_width));
    }

    return SVG_STATUS_SUCCESS;
}

/* The ellipse and arc functions below are:
 
   Copyright (C) 2000 Eazel, Inc.
  
   Author: Raph Levien <raph@artofcode.com>

   This is adapted from svg-path in Gill.
*/
static void
_svg_path_arc_segment (cairo_t *cr,
		       double xc, double yc,
		       double th0, double th1,
		       double rx, double ry, double x_axis_rotation)
{
    double sin_th, cos_th;
    double a00, a01, a10, a11;
    double x1, y1, x2, y2, x3, y3;
    double t;
    double th_half;

    sin_th = sin (x_axis_rotation * (M_PI / 180.0));
    cos_th = cos (x_axis_rotation * (M_PI / 180.0)); 
    /* inverse transform compared with rsvg_path_arc */
    a00 = cos_th * rx;
    a01 = -sin_th * ry;
    a10 = sin_th * rx;
    a11 = cos_th * ry;

    th_half = 0.5 * (th1 - th0);
    t = (8.0 / 3.0) * sin (th_half * 0.5) * sin (th_half * 0.5) / sin (th_half);
    x1 = xc + cos (th0) - t * sin (th0);
    y1 = yc + sin (th0) + t * cos (th0);
    x3 = xc + cos (th1);
    y3 = yc + sin (th1);
    x2 = x3 + t * sin (th1);
    y2 = y3 - t * cos (th1);

    cairo_curve_to (cr, a00 * x1 + a01 * y1, a10 * x1 + a11 * y1,
		a00 * x2 + a01 * y2, a10 * x2 + a11 * y2,
		a00 * x3 + a01 * y3, a10 * x3 + a11 * y3);
}

/**
 * _xsvg_path_arc_to: Add an arc to the given path
 *
 * rx: Radius in x direction (before rotation).
 * ry: Radius in y direction (before rotation).
 * x_axis_rotation: Rotation angle for axes.
 * large_arc_flag: 0 for arc length <= 180, 1 for arc >= 180.
 * sweep: 0 for "negative angle", 1 for "positive angle".
 * x: New x coordinate.
 * y: New y coordinate.
 *
 **/
static svg_status_t
_xsvg_arc_to (void		*closure,
	      double		rx,
	      double		ry,
	      double		x_axis_rotation,
	      int		large_arc_flag,
	      int		sweep_flag,
	      double		x,
	      double		y)
{
    xsvg_t *xsvg = closure;
    double sin_th, cos_th;
    double a00, a01, a10, a11;
    double x0, y0, x1, y1, xc, yc;
    double d, sfactor, sfactor_sq;
    double th0, th1, th_arc;
    int i, n_segs;
    double dx, dy, dx1, dy1, Pr1, Pr2, Px, Py, check;
    double curx, cury;

    cairo_current_point (xsvg->cr, &curx, &cury);

    sin_th = sin (x_axis_rotation * (M_PI / 180.0));
    cos_th = cos (x_axis_rotation * (M_PI / 180.0));

    dx = (curx - x) / 2.0;
    dy = (cury - y) / 2.0;
    dx1 =  cos_th * dx + sin_th * dy;
    dy1 = -sin_th * dx + cos_th * dy;
    Pr1 = rx * rx;
    Pr2 = ry * ry;
    Px = dx1 * dx1;
    Py = dy1 * dy1;
    /* Spec : check if radii are large enough */
    check = Px / Pr1 + Py / Pr2;
    if(check > 1)
    {
        rx = rx * sqrt(check);
        ry = ry * sqrt(check);
    }

    a00 = cos_th / rx;
    a01 = sin_th / rx;
    a10 = -sin_th / ry;
    a11 = cos_th / ry;
    x0 = a00 * curx + a01 * cury;
    y0 = a10 * curx + a11 * cury;
    x1 = a00 * x + a01 * y;
    y1 = a10 * x + a11 * y;
    /* (x0, y0) is current point in transformed coordinate space.
       (x1, y1) is new point in transformed coordinate space.
       
       The arc fits a unit-radius circle in this space.
    */
    d = (x1 - x0) * (x1 - x0) + (y1 - y0) * (y1 - y0);
    sfactor_sq = 1.0 / d - 0.25;
    if (sfactor_sq < 0) sfactor_sq = 0;
    sfactor = sqrt (sfactor_sq);
    if (sweep_flag == large_arc_flag) sfactor = -sfactor;
    xc = 0.5 * (x0 + x1) - sfactor * (y1 - y0);
    yc = 0.5 * (y0 + y1) + sfactor * (x1 - x0);
    /* (xc, yc) is center of the circle. */
    
    th0 = atan2 (y0 - yc, x0 - xc);
    th1 = atan2 (y1 - yc, x1 - xc);
    
    th_arc = th1 - th0;
    if (th_arc < 0 && sweep_flag)
	th_arc += 2 * M_PI;
    else if (th_arc > 0 && !sweep_flag)
	th_arc -= 2 * M_PI;

    /* XXX: I still need to evaluate the math performed in this
       function. The critical behavior desired is that the arc must be
       approximated within an arbitrary error tolerance, (which the
       user should be able to specify as well). I don't yet know the
       bounds of the error from the following computation of
       n_segs. Plus the "+ 0.001" looks just plain fishy. -cworth */
    n_segs = ceil (fabs (th_arc / (M_PI * 0.5 + 0.001)));
    
    for (i = 0; i < n_segs; i++) {
	_svg_path_arc_segment (xsvg->cr, xc, yc,
					th0 + i * th_arc / n_segs,
					th0 + (i + 1) * th_arc / n_segs,
					rx, ry, x_axis_rotation);
    }

    return _cairo_status_to_svg_status (cairo_status (xsvg->cr));
}


